# -*- coding: utf-8 -*-
# This file is part of modulemd-zanata
# Copyright (C) 2017-2018 Stephen Gallagher
#
# Fedora-License-Identifier: MIT
# SPDX-2.0-License-Identifier: MIT
# SPDX-3.0-License-Identifier: MIT
#
# This program is free software.
# For more information on the license, see COPYING.
# For more information on free software, see
# <https://www.gnu.org/philosophy/free-sw.en.html>.

import requests

KOJI_URL = 'https://koji.fedoraproject.org/kojihub'
ZANATA_URL = "https://fedora.zanata.org"
ZANATA_PROJECT = "fedora-modularity-translations"
ZANATA_DOCUMENT = "fedora-modularity-translations"


def get_fedora_rawhide_version(session, debug=False):
    # Koji sometimes disconnects for no apparent reason. Retry up to 5
    # times before failing.
    for attempt in range(5):
        try:
            build_targets = session.getBuildTargets('rawhide')
        except requests.exceptions.ConnectionError:
            if debug:
                print("Connection lost while retriving rawhide branch, "
                      "retrying...",
                      file=sys.stderr)
        else:
            # Succeeded this time, so break out of the loop
            break

    return build_targets[0][
        'build_tag_name'].partition('-build')[0]


def get_tags_for_fedora_branch(branch):
    return ['%s-modular' % branch,
            '%s-modular-override' % branch,
            '%s-modular-pending' % branch,
            '%s-modular-signing-pending' % branch,
            '%s-modular-updates' % branch,
            '%s-modular-updates-candidate' % branch,
            '%s-modular-updates-pending' % branch,
            '%s-modular-updates-testing' % branch,
            '%s-modular-updates-testing-pending' % branch]



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# This file is part of modulemd-zanata
# Copyright (C) 2017-2018 Stephen Gallagher
#
# Fedora-License-Identifier: MIT
# SPDX-2.0-License-Identifier: MIT
# SPDX-3.0-License-Identifier: MIT
#
# This program is free software.
# For more information on the license, see COPYING.
# For more information on free software, see
# <https://www.gnu.org/philosophy/free-sw.en.html>.

import click
import gi
import koji
import mmdzanata
import mmdzanata.fedora
import subprocess
import sys

from babel.messages import pofile

gi.require_version('Modulemd', '1.0')
from gi.repository import Modulemd


##############################################################################
# Common options for all commands                                            #
##############################################################################


@click.group()
@click.option('--debug/--no-debug', default=False)
@click.option('-k', '--koji-url',
              default=mmdzanata.fedora.KOJI_URL,
              type=str, help="The URL of the Koji build system.",
              show_default=True,
              metavar="<URL>")
@click.option('-b', '--branch', default="rawhide", type=str,
              help="The distribution release",
              metavar="<branch_name>")
@click.option('-z', '--zanata-url',
              default=mmdzanata.fedora.ZANATA_URL,
              type=str, help="The Zanata URL",
              show_default=True,
              metavar="<zanata_project>")
@click.option('-p', '--zanata-project',
              default=mmdzanata.fedora.ZANATA_PROJECT,
              type=str, help="The Zanata project",
              show_default=True,
              metavar="<zanata_project>")
@click.option('-f', '--zanata-translation-document',
              default=mmdzanata.fedora.ZANATA_DOCUMENT,
              help="The name of the translated file in Zanata.",
              show_default=True,
              metavar="<translation_document>")
@click.pass_context
def cli(ctx, debug, branch, koji_url, zanata_url, zanata_project,
        zanata_translation_document):

    ctx.obj = dict()
    ctx.obj['debug'] = debug

    ctx.obj['session'] = koji.ClientSession(koji_url)

    ctx.obj['branch'] = branch

    if branch == "rawhide":
        ctx.obj['branch'] = mmdzanata.fedora.get_fedora_rawhide_version(
            ctx.obj['session'])

    ctx.obj['zanata_url'] = zanata_url
    ctx.obj['zanata_project'] = zanata_project
    ctx.obj['zanata_translation_document'] = zanata_translation_document

##############################################################################
# Subcommands                                                                #
##############################################################################

##############################################################################
# `mmdzanata extract`                                                        #
##############################################################################


@cli.command()
@click.option('--upload/--no-upload', default=True,
              help='Whether to automatically push extracted strings to '
              'Zanata',
              show_default=True)
@click.pass_context
def extract(ctx, upload):
    """
    Extract translations from all modules included in a particular version of
    Fedora or EPEL.
    """

    catalog = mmdzanata.get_module_catalog_from_tags(
        ctx.parent.obj['session'], mmdzanata.fedora.get_tags_for_fedora_branch(
            ctx.parent.obj['branch']),
        debug=ctx.parent.obj['debug'])

    potfile = "%s.pot" % ctx.parent.obj['zanata_translation_document']

    with open(potfile, mode="wb") as f:
        pofile.write_po(f, catalog, sort_by_file=True)

    print("Wrote extracted strings for %s to %s" % (ctx.obj['branch'],
                                                    potfile))

    # Optionally upload the extracted strings directly to Zanata
    if upload:
        # Use the zanata-cli to upload the pot file
        # It would be better to use the REST API directly here, but the XML
        # payload format is not documented.
        zanata_args = ['/usr/bin/zanata-cli', '-B', '-e', 'push',
                       '--url', ctx.parent.obj['zanata_url'],
                       '--project', ctx.parent.obj['zanata_project'],
                       '--project-type', 'gettext',
                       '--project-version', ctx.parent.obj['branch']]
        result = subprocess.run(zanata_args, capture_output=True)
        if result.returncode:
            print(result.stderr)
            print(result.stdout)
            sys.exit(1)


##############################################################################
# `mmdzanata generate_metadata`                                              #
##############################################################################

@cli.command()
@click.pass_context
def generate_metadata(ctx):
    """
    :return: 0 on successful creation of modulemd-translation,
    nonzero on failure.
    """

    zanata_rest_url = "%s/rest" % ctx.parent.obj['zanata_url']

    translations = mmdzanata.get_modulemd_translations(
        zanata_rest_url,
        ctx.parent.obj['zanata_project'],
        ctx.parent.obj['branch'],
        ctx.parent.obj['zanata_translation_document'],
        ctx.parent.obj['debug']
    )

    Modulemd.dump(sorted(translations), "%s.yaml" % (
        ctx.parent.obj['zanata_translation_document']))

    print("Wrote modulemd-translations YAML to %s.yaml" % (
        ctx.parent.obj['zanata_translation_document']))


if __name__ == "__main__":
    cli(obj={})



# -*- coding: utf-8 -*-
# This file is part of modulemd-zanata
# Copyright (C) 2017-2018 Stephen Gallagher
#
# Fedora-License-Identifier: MIT
# SPDX-2.0-License-Identifier: MIT
# SPDX-3.0-License-Identifier: MIT
#
# This program is free software.
# For more information on the license, see COPYING.
# For more information on free software, see
# <https://www.gnu.org/philosophy/free-sw.en.html>.
#
# This module provides utility functions for interacting with Zanata
# translations of Fedora-style modules.

import sys
import gi
import requests

from babel.messages import Catalog, pofile
from datetime import datetime
from io import BytesIO

gi.require_version('Modulemd', '1.0')
from gi.repository import Modulemd


##############################################################################
# Exceptions                                                                 #
##############################################################################


class MmdZanataError(Exception):
    pass


class NonexistentProjectError(MmdZanataError):

    def __init__(self, project, version):
        """
        Exception thrown when the user has requested a Zanata project that
        does not exist
        :param project: The nonexistent project name
        :param version: The version of the project in Zanata
        """
        self.project = project
        self.version = version
        self.message = "The project %s does not exist in Zanata" % project


class UnexpectedHTTPResponse(MmdZanataError):

    def __init__(self, status_code, body):
        """
        Exception thrown when an unexpected response is received from an HTTP
        request.
        :param body: The response body
        """
        self.status_code = status_code
        self.body = body


def get_latest_modules_in_tag(session, tag, debug=False):
    """
    Get the most-recently built versions of each (module,stream) pair from
    a Koji tag
    :param session: A Koji session
    :param tag: A koji tag
    :return: A list of the most recent build of all modules in the tag.
    """

    # Koji sometimes disconnects for no apparent reason. Retry up to 5
    # times before failing.
    for attempt in range(5):
        try:
            tagged = session.listTagged(tag, latest=False)
        except requests.exceptions.ConnectionError:
            if debug:
                print("Connection lost while retrieving builds for tag %s, "
                      "retrying..." % tag,
                      file=sys.stderr)
        else:
            # Succeeded this time, so break out of the loop
            break

    # Find the latest, in module terms.  Pungi does this.
    # Collect all contexts that share the same NSV.
    NSVs = {}
    for entry in tagged:
        name, stream = entry['name'], entry['version']
        version = entry['release'].rsplit('.', 1)[0]

        NSVs[name] = NSVs.get(name, {})
        NSVs[name][stream] = NSVs[name].get(stream, {})
        NSVs[name][stream][version] = NSVs[name][stream].get(version, [])
        NSVs[name][stream][version].append(entry)

    latest = []
    for name in NSVs:
        for stream in NSVs[name]:
            version = sorted(list(NSVs[name][stream].keys()))[-1]
            latest.extend(NSVs[name][stream][version])

    return latest


def get_module_catalog_from_tags(session, tags, debug=False):
    """
    Construct a Babel translation source catalog from the contents of the
    provided tags.
    :param session: A Koji session
    :param tags: A set of Koji tags from which module metadata should be pulled
    :param debug: Whether to print debugging information to the console
    :return: A babel.messages.Catalog containing extracted translatable strings
    from any modules in the provided tags. Raises an exception if any of the
    retrieved modulemd is invalid.
    """

    catalog = Catalog(project="fedora-modularity-translations")

    tagged_builds = []
    for tag in tags:
        tagged_builds.extend(get_latest_modules_in_tag(session, tag, debug))

    # Make the list unique since some modules may have multiple tags
    unique_builds = {}
    for build in tagged_builds:
        unique_builds[build['id']] = build

    for build_id in unique_builds.keys():
        # Koji sometimes disconnects for no apparent reason. Retry up to 5
        # times before failing.
        for attempt in range(5):
            try:
                build = session.getBuild(build_id)
            except requests.exceptions.ConnectionError:
                if debug:
                    print("Connection lost while processing buildId %s, "
                          "retrying..." % build_id,
                          file=sys.stderr)
            else:
                # Succeeded this time, so break out of the loop
                break
        if debug:
            print("Processing %s:%s" % (build['package_name'], build['nvr']))

        modulemds = Modulemd.objects_from_string(
            build['extra']['typeinfo']['module']['modulemd_str'])

        # We should only get a single modulemd document from Koji
        if len(modulemds) != 1:
            raise ValueError("Koji build %s returned multiple modulemd YAML "
                             "documents." % build['nvr'])

        mmd = modulemds[0]

        # Process the summary
        msg = catalog.get(mmd.props.summary)
        if msg:
            locations = msg.locations
        else:
            locations = []
        locations.append(("%s;%s;summary" % (
            mmd.props.name, mmd.props.stream), 1))
        catalog.add(mmd.props.summary, locations=locations)

        # Process the description
        msg = catalog.get(mmd.props.description)
        if msg:
            locations = msg.locations
        else:
            locations = []
        locations.append(("%s;%s;description" % (
            mmd.props.name, mmd.props.stream), 2))
        catalog.add(mmd.props.description, locations=locations)

        # Get any profile descriptions
        for profile_name, profile in modulemds[0].peek_profiles().items():
            if profile.props.description:
                msg = catalog.get(profile.props.description)
                if msg:
                    locations = msg.locations
                else:
                    locations = []

                locations.append(("%s;%s;profile;%s" % (
                    mmd.props.name,
                    mmd.props.stream,
                    profile.props.name),
                    3))
                catalog.add(profile.props.description, locations=locations)

    return catalog


def get_translated_locales(zanata_rest_url, zanata_project,
                           zanata_project_version,
                           debug=False):
    # Get the statistics on the translation project for this version
    stats_url = zanata_rest_url + "/stats/proj/%s/iter/%s" % (
        zanata_project, zanata_project_version)

    r = requests.get(stats_url, headers={"Accept": "application/json"})
    if r.status_code == 404:
        print("Project '%s:%s' does not exist." % (
            zanata_project, zanata_project_version),
            file=sys.stderr)
        raise NonexistentProjectError(zanata_project, zanata_project_version)
    elif r.status_code != 200:
        raise UnexpectedHTTPResponse(r.status_code, r.content)

    # We will pull down information for any locale that is at least partially
    # translated
    translated_locales = [t["locale"]
                          for t in r.json()['stats']
                          if t["translated"] > 0]

    if debug:
        print("Available locales: ")
        for locale in translated_locales:
            print("* %s" % locale)

    return translated_locales


def get_modulemd_translations_from_catalog_dict(catalog_dict):
    now = datetime.utcnow()
    modified = int("%04d%02d%02d%02d%02d%02d" % (
        now.date().year,
        now.date().month,
        now.date().day,
        now.time().hour,
        now.time().minute,
        now.time().second
    ))

    # Translation entries keyed by name, stream and locale
    entries = dict()

    mmd_translations = dict()
    for locale, catalog in catalog_dict.items():
        for msg in catalog:
            if not msg.locations or not msg.string:
                # Skip any message that doesn't actually contain a message
                continue

            for location, _ in msg.locations:
                split_location = location.split(';')
                if len(split_location) < 3 or len(split_location) > 5:
                    print("Invalid location clue in translation data: %s" % (
                        location), file=sys.stderr)

                module_name = split_location[0]
                module_stream = split_location[1]

                try:
                    entry = entries[(module_name, module_stream, locale)]
                except KeyError:
                    entry = Modulemd.TranslationEntry.new(locale)

                # Summary Translation
                if split_location[2] == "summary":
                    entry.set_summary(msg.string)

                # Description Translation
                elif split_location[2] == "description":
                    entry.set_description(msg.string)

                # Translation of profile descriptions
                elif split_location[2] == "profile":
                    entry.set_profile_description(split_location[3],
                                                  msg.string)

                entries[(module_name, module_stream, locale)] = entry

    for (module_name, module_stream, locale), entry in entries.items():
        # Validate that the translation entry has both summary and description
        # which are mandatory.
        if not entry.get_summary() or not entry.get_description():
            continue

        # Otherwise, add or update the translation for this module and stream
        try:
            mmdtranslation = mmd_translations[(module_name,
                                               module_stream)]
        except KeyError:
            mmdtranslation = Modulemd.Translation.new_full(
                module_name=module_name,
                module_stream=module_stream,
                mdversion=1,
                modified=modified
            )

        mmdtranslation.add_entry(entry)
        mmd_translations[(module_name, module_stream)] = mmdtranslation

    return mmd_translations.values()


def get_modulemd_translations(zanata_rest_url, zanata_project,
                              os_branch, zanata_translation_file,
                              debug=False):
    translated_locales = get_translated_locales(zanata_rest_url,
                                                zanata_project,
                                                os_branch,
                                                debug)

    catalogs = dict()
    for loc in translated_locales:
        # Get the translation data for this locale
        pofile_url = zanata_rest_url + \
            "/file/translation/%s/%s/%s/po?docId=%s" % (
                zanata_project, os_branch, loc,
                zanata_translation_file)
        r = requests.get(pofile_url,
                         headers={"Accept": "application/octet-stream"})
        if r.status_code != 200:
            print("Could not retrieve translations for %s" % loc,
                  file=sys.stderr)
            continue

        if debug:
            print("PO content for locale '%s'" % loc)
            print(r.text)

        # Read the po file into a catalog, indexed by the locale
        catalogs[loc] = pofile.read_po(
            BytesIO(r.content),
            domain="fedora-modularity-translations")

    translations = get_modulemd_translations_from_catalog_dict(catalogs)
    if debug:
        for translation in translations:
            print(translation.dumps())

    return translations



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# This file is part of mmdzanata
# Copyright (C) 2017-2018 Stephen Gallagher
#
# Fedora-License-Identifier: MIT
# SPDX-2.0-License-Identifier: MIT
# SPDX-3.0-License-Identifier: MIT
#
# This program is free software.
# For more information on the license, see COPYING.
# For more information on free software, see
# <https://www.gnu.org/philosophy/free-sw.en.html>.

from setuptools import setup

setup(
    name='mmdzanata',
    version='0.1',
    packages=['mmdzanata', 'mmdzanata.fedora'],
    url='',
    license='MIT',
    author='Stephen Gallagher',
    author_email='sgallagh@redhat.com',
    description='Tools for working with translations of modulemd',
    # mmdzanata also requires koji, libmodulemd and zanata-client which are not
    # available on PyPI and must be installed separately. On Fedora, this is
    # done with `dnf install koji libmodulemd zanata-client`
    install_requires=[
        'click',
        'requests',
        'babel',
    ],
    entry_points={
        'console_scripts': ['mmdzanata=mmdzanata.cli:cli'],
    },
)
